#include "framebuffer.h"
#include "v3.h"
#include "math.h"
#include <iostream>
#include <fstream>
#include <float.h>
#include "scene.h"
#include <tiffio.h>

FrameBuffer::FrameBuffer(int u0, int v0, 
  int _w, int _h) : Fl_Gl_Window(u0, v0, _w, _h, 0) {

  w = _w;
  h = _h;
  pix = new unsigned int[w*h];
  zb = new float[w*h];
  FrameSetup();

}


void FrameBuffer::FrameSetup() {

  for (int i = 0; i < w*h; i++) {
    zb[i] = 0.0f;
  }
  Set(0xFFFFFFFF);

}

void FrameBuffer::draw() {

    glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);

}

void FrameBuffer::Set(unsigned int color) {

  for (int i = 0; i < w*h; i++)
    pix[i] = color;

}

void FrameBuffer::Set(int u, int v, unsigned int color) {

  pix[(h-1-v)*w+u] = color;

}

void FrameBuffer::SetGuarded(int u, int v, unsigned int color) {

  if (u < 0 || u > w-1 || v < 0 || v > h-1)
    return;

  Set(u, v, color);

}

unsigned int FrameBuffer::Get(int u, int v) {

  return pix[(h-1-v)*w+u];

}

int FrameBuffer::handle(int event)  {  

  switch(event) 
  {   
  case FL_KEYBOARD: {
    KeyboardHandle();
    return 0;
  }
  default:
    break;
  }
  return 0;

}


void FrameBuffer::KeyboardHandle() {

  float tstep = 3.0f;
  float rstep = 3.0f;
  int key = Fl::event_key();
  switch (key) {
    case '.': {
      scene->ppc->TranslateRight(tstep);
      scene->Render();
      break;
    }
    case ',': {
      scene->ppc->TranslateRight(-tstep);
      scene->Render();
      break;
    }
    case '1': {
      scene->ppc->TranslateUp(tstep);
      scene->Render();
      break;
    }
    case '2': {
      scene->ppc->TranslateUp(-tstep);
      scene->Render();
      break;
    }
    case FL_Up: {
      scene->ppc->TranslateForward(tstep);
      scene->Render();
      break;
    }
    case FL_Down: {
      scene->ppc->TranslateForward(-tstep);
      scene->Render();
      break;
    }
    case FL_Left: {
      scene->ppc->RotateAboutAxisThroughC(scene->ppc->b.Normalized()*-1.0f, rstep);
      scene->Render();
      break;
    }
    case FL_Right: {
      scene->ppc->RotateAboutAxisThroughC(scene->ppc->b.Normalized()*-1.0f, -rstep);
      scene->Render();
      break;
    }
    case 'q': {
      scene->ppc->RotateAboutAxisThroughC(
        scene->ppc->a.Normalized(), rstep);
      scene->Render();
      break;
    }
    case 'w': {
      scene->ppc->RotateAboutAxisThroughC(
        scene->ppc->a.Normalized(), -rstep);
      scene->Render();
      break;
    }
    case 'z': {
      scene->ppc->ZoomIn(1.1f);
      scene->Render();
      break;
    }
    case 'x': {
      scene->ppc->ZoomIn(0.9f);
      scene->Render();
      break;
    }
    case 'a': {
      cerr << "INFO: pressed a" << endl;
      break;
    }
    default:
      cerr << "INFO: do not understand keypress" << endl;
  }
}

void FrameBuffer::MakeChecker(int cw, unsigned int col0, 
  unsigned int col1) {

  for (int v = 0; v < h; v++) {
    for (int u = 0; u < w; u++) {
      int cu = u / cw;
      int cv = v / cw;
      if ((cu + cv) % 2)
        Set(u, v, col0);
      else
        Set(u, v, col1);
    }
  }


}

void FrameBuffer::DrawPoint3D(V3 P, int psize, 
                              unsigned int pcol, PPC *ppc) {

  V3 projP;
  if (!ppc->Project(P, projP))
    return;

  int ui = (int) projP[0];
  int vi = (int) projP[1];

  for (int u = ui - psize/2; u <= ui + psize/2; u++) {
    for (int v = vi - psize/2; v <= vi + psize/2; v++) {
      if (u < 0 || u > w-1 || v < 0 || v > h-1)
        continue;
      Set(u, v, pcol);
    }
  }

}

void FrameBuffer::DrawSegment3D(V3 P0, 
  V3 P1, V3 c0, V3 c1, PPC *ppc) {


    V3 projP0, projP1;
    if (!ppc->Project(P0, projP0)) {
      if (projP0[0] == FLT_MAX)
        return;
    }
    if (!ppc->Project(P1, projP1)) {
      if (projP1[0] == FLT_MAX)
        return;
    }

    DrawSegment2D(projP0, projP1, c0, c1);
  
}

void FrameBuffer::DrawSegment3D(V3 P0, 
  V3 P1, unsigned int scol, PPC *ppc) {

    V3 c0;
    c0.SetFromColor(scol);
    V3 c1 = c0;
    DrawSegment3D(P0, P1, c0, c1, ppc);

}

void FrameBuffer::DrawSegment2D(V3 P0, V3 P1, 
  V3 c0, V3 c1) {

  float dx = fabsf(P0[0]-P1[0]);
  float dy = fabsf(P0[1]-P1[1]);

  int stepsN;
  if (dx > dy)
    stepsN = (int) ceilf(dx);
  else
    stepsN = (int) ceilf(dy);


  stepsN++;
  V3 currP, currc;
  V3 dP = (P1-P0) / (float) (stepsN-1);
  V3 dc = (c1-c0) / (float) (stepsN-1);
  int i;
  for (i = 0, currP = P0, currc = c0; i < stepsN; i++, 
    currP = currP + dP, currc = currc + dc) {
    int ui = (int) currP[0];    
    int vi = (int) currP[1];
    if (IsCloserThenSet(ui, vi, currP[2])) {
      Set(ui, vi, currc.GetColor());
    }
  }

}

bool FrameBuffer::IsCloserThenSet(int ui, int vi, float z) {

  if (ui < 0 || vi < 0 || ui > w-1 || vi > h-1)
    return false;

  if (zb[(h-1-vi)*w+ui] > z)
    return false;

  zb[(h-1-vi)*w+ui] = z;
  return true;

}

void FrameBuffer::Save(char *fname) {
  
  TIFF* out = TIFFOpen(fname, "w");
  int sampleperpixel = 4;

  if(out == NULL) {
    cerr << fname << " could not be opened" << endl;
    return;
  }
  
  TIFFSetField(out, TIFFTAG_IMAGEWIDTH, w);
  TIFFSetField(out, TIFFTAG_IMAGELENGTH, h); 
  TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, sampleperpixel);
  TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
  TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  
  TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  TIFFSetField(out, TIFFTAG_SOFTWARE, "Tiff writer");
  TIFFSetField(out, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
  
  
  // We set the strip size of the file to be size of one row of pixels
  TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(out, w*sampleperpixel));
  
  // Now writing image to the file one strip at a time
  for (int row = 0; row < h; row++) {
    
    if (TIFFWriteScanline(out, &pix[(h-1-row)*w], row) < 0)
      break;
  }
  
  TIFFClose(out);
  
}

int FrameBuffer::Load(char *tiffFileName) {
  
  if (!strcmp(tiffFileName + strlen(tiffFileName)-4, ".tif")) {
    
    ifstream tst(tiffFileName);
    if (tst.fail()) {
      cerr << "INFO: cannot open: " << tiffFileName << endl;
      return 0;
    }
    tst.close();
    
    TIFF* tif = TIFFOpen(tiffFileName, "r");
    
    if (tif) {
      if (pix)
        delete pix;
      
      cerr << "INFO: started loading: " << tiffFileName << endl;
      
      TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
      TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
      pix = new unsigned int[w*h];
      
      if (!TIFFReadRGBAImage(tif, w, h, (uint32*) pix, 0)) {
        delete pix;
        pix = 0;
        w = h = 0;
        return 0;
      }
      
      TIFFClose(tif);
      label(tiffFileName);
      cerr << "INFO: done loading: " << tiffFileName << endl;
      return 1;
    }
    return 0;
  }
  cerr << "INFO: do not know how to open *" << tiffFileName + strlen(tiffFileName)-4 << endl;
  return 0;
}


bool FrameBuffer::IsInShadow(V3 projP) {

  if (
    projP[0] < 0.0f || projP[0] >= (float) w ||
    projP[1] < 0.0f || projP[1] >= (float) h)

    return false;
  int uv = (h-1-(int)projP[1])*w+(int)projP[0];

  return (projP[2] + 0.05f <  zb[uv]);

}